from advanced_alchemy.base import UUIDAuditBase
from sqlalchemy.orm import Mapped, mapped_column, relationship


class RoleModel(UUIDAuditBase):
    __tablename__ = "roles"

    name: Mapped[str] = mapped_column(unique=True)
    users = relationship(
        "UserModel",
        secondary="users_roles",
        back_populates="roles",
        lazy="selectin",
    )



from typing import Any
from uuid import UUID

from litestar import Controller, Request, get, post
from litestar.datastructures import State
from litestar.di import Provide
from litestar.exceptions import HTTPException
from litestar.status_codes import HTTP_400_BAD_REQUEST
import msgspec

from auth.services import (
    AuthService,
    AccountDTO,
    LoginRequestDTO,
    LogoutRequestDTO,
    SignUpRequestDTO,
    UserNotEnabledError,
    provide_auth_service,
    provide_user_service,
    InvalidEmailError,
    InvalidPasswordError,
)
from auth.services.auth_service import AccountUserDTO


class SignUpRequestScheme(msgspec.Struct):
    email: str
    password: str


class SignUpResponseScheme(msgspec.Struct):
    message: str


class LoginResponseScheme(AccountDTO): ...


class ActivateAccountResponseScheme(AccountDTO): ...


class LoginRequestScheme(msgspec.Struct):
    email: str
    password: str


class ActivateAccountRequestScheme(msgspec.Struct):
    code: str


class AccountMeResponseScheme(msgspec.Struct):
    session_id: UUID
    user: AccountUserDTO


class AccountController(Controller):
    path = "/account"

    dependencies = {
        "user_service": Provide(provide_user_service),
        "auth_service": Provide(provide_auth_service),
    }

    @get("/me")
    async def account(self, request: Request[AccountDTO, Any, State]) -> AccountMeResponseScheme:
        return AccountMeResponseScheme(session_id=request.user.session_id, user=request.user.user)

    @post("/login", exclude_from_auth=True)
    async def login(self, request: Request, data: LoginRequestScheme, auth_service: AuthService) -> LoginResponseScheme:
        device = request.headers.get("User-Agent")
        try:
            login_user = await auth_service.login(
                LoginRequestDTO(
                    email=data.email,
                    password=data.password,
                    device=device,
                )
            )
        except UserNotEnabledError:
            raise HTTPException(f"User {data.email} is not enabled", status_code=HTTP_400_BAD_REQUEST)
        except InvalidEmailError:
            raise HTTPException(f"User with email {data.email} not found", status_code=HTTP_400_BAD_REQUEST)
        except InvalidPasswordError:
            raise HTTPException("Invalid password", status_code=HTTP_400_BAD_REQUEST)

        return LoginResponseScheme(
            token=login_user.token,
            session_id=login_user.session_id,
            user=login_user.user,
        )

    @post("/signup", exclude_from_auth=True)
    async def sign_up(self, data: SignUpRequestScheme, auth_service: AuthService) -> SignUpResponseScheme:
        await auth_service.signup(SignUpRequestDTO(email=data.email, password=data.password))
        return SignUpResponseScheme(message="success")

    @post("/logout")
    async def logout(self, auth_service: AuthService, request: Request[AccountDTO, Any, State]) -> None:
        await auth_service.logout(
            LogoutRequestDTO(
                user_id=request.user.user.id,
                session_id=request.user.session_id,
                device=request.headers.get("User-Agent"),
            )
        )

    @post("/activate", exclude_from_auth=True)
    async def activate(
        self,
        data: ActivateAccountRequestScheme,
        auth_service: AuthService,
        request: Request,
    ) -> ActivateAccountResponseScheme:
        device = request.headers.get("User-Agent")
        activated_account = await auth_service.verify_user_email(data.code, device)

        return ActivateAccountResponseScheme(
            token=activated_account.token,
            session_id=activated_account.session_id,
            user=activated_account.user,
        )



from typing import AsyncGenerator
from uuid import UUID
from litestar.exceptions import NotAuthorizedException

import msgspec
import pytest

from auth.factories import UserFactory
from auth.services import (
    AuthService,
    InvalidEmailError,
    InvalidPasswordError,
    LoginRequestDTO,
    LogoutRequestDTO,
    SignUpRequestDTO,
    UserNotEnabledError,
    UserEmailNotVerifiedError,
    UserService,
    WrongAuthCodeError,
    AccountDTO,
)
from tests.integrations import TestMailClient


class AuthInfoFixture(msgspec.Struct):
    user_id: UUID
    email: str
    code: str
    password: str


@pytest.fixture
async def auth_info(auth_service: AuthService) -> AsyncGenerator[AuthInfoFixture]:
    user = UserFactory.build()
    auth_code = await auth_service.signup(SignUpRequestDTO(email=user.email, password=user.password))
    yield AuthInfoFixture(
        user_id=auth_code.user_id,
        email=user.email,
        password=user.password,
        code=auth_code.code,
    )
    await auth_service.user_service.delete(auth_code.user_id)


async def test_auth_signup(auth_service: AuthService, mail_client: TestMailClient, user_service: UserService):
    user = UserFactory.build()
    auth_code = await auth_service.signup(SignUpRequestDTO(email=user.email, password=user.password))
    assert mail_client.messages == [([user.email], "Sign up", auth_code.code)]

    created_user = await user_service.get(auth_code.user_id)
    assert created_user.email == user.email

    # teardown
    await auth_service.user_service.delete(auth_code.user_id)


async def test_auth_login_success(auth_service: AuthService, auth_info: AuthInfoFixture):
    activated_user = await auth_service.verify_user_email(auth_info.code, "testclient")
    account_session = await auth_service.login(
        LoginRequestDTO(
            email=auth_info.email,
            password=auth_info.password,
            device="testclient",
        )
    )
    assert account_session.user.id == activated_user.user.id


async def test_auth_login_wrong_auth_code(auth_info: AuthInfoFixture, auth_service: AuthService):
    with pytest.raises(WrongAuthCodeError):
        await auth_service.verify_user_email("12345678", "testclient")


async def test_auth_login_user_email_not_verified(auth_info: AuthInfoFixture, auth_service: AuthService):
    with pytest.raises(UserEmailNotVerifiedError):
        await auth_service.login(
            LoginRequestDTO(email=auth_info.email, password=auth_info.password, device="testclient")
        )


async def test_auth_login_wrong_email(auth_service: AuthService, auth_info: AuthInfoFixture):
    await auth_service.verify_user_email(auth_info.code, "testclient")

    with pytest.raises(InvalidEmailError):
        await auth_service.login(
            LoginRequestDTO(email="wrong@mail.com", password=auth_info.password, device="testclient")
        )


async def test_auth_login_wrong_password(auth_service: AuthService, auth_info: AuthInfoFixture):
    await auth_service.verify_user_email(auth_info.code, "testclient")

    with pytest.raises(InvalidPasswordError):
        await auth_service.login(LoginRequestDTO(email=auth_info.email, password="wrong_password", device="testclient"))


async def test_auth_get_account(auth_service: AuthService, auth_info: AuthInfoFixture):
    activated_user = await auth_service.verify_user_email(auth_info.code, "testclient")
    account_session = await auth_service.login(
        LoginRequestDTO(
            email=auth_info.email,
            password=auth_info.password,
            device="testclient",
        )
    )
    assert account_session.user.id == activated_user.user.id
    account = await auth_service.get_account(account_session.token)
    assert account.user.id == activated_user.user.id


async def test_auth_logout_success(auth_service: AuthService, auth_info: AuthInfoFixture):
    activated_user = await auth_service.verify_user_email(auth_info.code, "testclient")
    account_session = await auth_service.login(
        LoginRequestDTO(
            email=auth_info.email,
            password=auth_info.password,
            device="testclient",
        )
    )
    account = await auth_service.get_account(account_session.token)

    await auth_service.logout(
        LogoutRequestDTO(
            user_id=activated_user.user.id,
            device="testclient",
            session_id=account_session.session_id,
        )
    )

    with pytest.raises(NotAuthorizedException):
        await auth_service.check_session(
            account,
            "testclient",
        )



from pathlib import Path

import click
from click import Group
from litestar.plugins import CLIPlugin
import yaml

from core.cli import coro
from core.database import session_maker
from core.mail import MailClient
from core.settings import settings

from .loaders import AuthLoader
from .services import provide_permission_service, provide_role_service, provide_user_service


class AuthPlugin(CLIPlugin):
    def on_cli_init(self, cli: Group) -> None:
        @cli.group(help="Manage auth, load data with ``load`` command")
        @click.version_option(prog_name="auth")
        def auth(): ...

        @auth.command(help="load auth data")
        @coro
        async def load():
            async with session_maker() as session:
                click.echo("Loading auth data... ")
                loader = AuthLoader(
                    user_service=await provide_user_service(session),
                    role_service=await provide_role_service(session),
                    permission_service=await provide_permission_service(session),
                )

                with open(f"{Path(__file__).parent.resolve()}/fixtures.yaml") as f:
                    data = yaml.safe_load(f)
                    await loader.load(data)

        @auth.command(help="load auth data")
        @coro
        async def clear():
            async with session_maker() as session:
                user_service = await provide_user_service(session)
                await user_service.delete_where()
                click.echo("Clear auth data")

        @auth.command(help="Send test mail")
        @click.argument("recipient")
        def send_mail(recipient):
            mail_controller = MailClient(settings.mail_config)
            mail_controller.send([recipient], "test", "test")



from advanced_alchemy.base import UUIDAuditBase
from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column


class AuthCodeModel(UUIDAuditBase):
    __tablename__ = "auth_codes"

    code: Mapped[str] = mapped_column()
    user_id = mapped_column(ForeignKey("users.id", ondelete="CASCADE"), unique=True)



from litestar import Controller, get
import msgspec

from auth.guards import is_debug_guard


class E2EActivateCodeSchemeResponse(msgspec.Struct):
    code: str

class E2EController(Controller):
    guards = [is_debug_guard]
    path = "e2e"

    @get('/activate-code')
    async def get_activate_code(self) -> E2EActivateCodeSchemeResponse:
        return E2EActivateCodeSchemeResponse(code='super code')



from litestar import Router

from .account_controller import AccountController
from .permission_controller import PermissionController
from .role_controller import RoleController
from .user_controller import UserController

auth_router = Router(
    path="/api/auth",
    route_handlers=[
        UserController,
        RoleController,
        PermissionController,
        AccountController,
    ],
)

__all__ = ["auth_router"]



from typing import Any

from auth.models import UserModel
from auth.services import AuthService, PermissionService, RoleService, UserService


class AuthLoader:
    def __init__(
        self,
        user_service: UserService,
        role_service: RoleService,
        permission_service: PermissionService,
    ):
        self.user_service = user_service
        self.role_service = role_service
        self.permission_service = permission_service

    async def load(self, data: dict[str, Any]):
        await self.user_service.create_many(
            [
                UserModel(
                    email=user["email"],
                    password=AuthService.hash_password(user["password"]),
                    is_enabled=True,
                    is_email_verified=True,
                )
                for user in data.get("users", [])
            ],
            auto_commit=True,
        )



from litestar.status_codes import HTTP_200_OK, HTTP_201_CREATED, HTTP_401_UNAUTHORIZED, HTTP_400_BAD_REQUEST
from litestar.testing import AsyncTestClient
import msgspec
from auth.factories import UserFactory
from auth.models import UserModel, AuthCodeModel
from auth.services import API_KEY_HEADER, TOKEN_PREFIX, AuthService, LoginRequestDTO, SignUpRequestDTO
import pytest_asyncio


class ActivatedDataFixture(msgspec.Struct):
    user: UserModel
    auth_code: AuthCodeModel


@pytest_asyncio.fixture
async def activated_data(auth_service: AuthService) -> ActivatedDataFixture:
    user = UserFactory.build()
    auth_code = await auth_service.signup(SignUpRequestDTO(email=user.email, password=user.password))
    await auth_service.verify_user_email(auth_code.code, "testclient")

    return ActivatedDataFixture(user=user, auth_code=auth_code)


async def test_account_signup(http_api_client):
    response = await http_api_client.post(
        "/api/auth/account/signup",
        json={
            "email": "newuser@mail.com",
            "password": "newuserpassword",
        },
    )
    assert response.status_code == HTTP_201_CREATED


async def test_account_login(activated_data, http_api_client: AsyncTestClient):
    login_response = await http_api_client.post(
        "/api/auth/account/login",
        json={"email": activated_data.user.email, "password": activated_data.user.password},
    )
    assert login_response.status_code == HTTP_201_CREATED
    login_response_json = login_response.json()
    assert login_response_json["user"]["email"] == activated_data.user.email
    assert login_response_json["user"]["id"] == str(activated_data.auth_code.user_id)


async def test_account_login_wrong_email(activated_data, http_api_client: AsyncTestClient):
    wrong_email = "wrong@email.com"

    login_response = await http_api_client.post(
        "/api/auth/account/login",
        json={"email": wrong_email, "password": activated_data.user.password},
    )
    assert login_response.status_code == HTTP_400_BAD_REQUEST
    assert login_response.json()["detail"] == f"User with email {wrong_email} not found"


async def test_account_login_wrong_password(activated_data, http_api_client: AsyncTestClient):
    login_response = await http_api_client.post(
        "/api/auth/account/login",
        json={"email": activated_data.user.email, "password": "wrong_password"},
    )
    assert login_response.status_code == HTTP_400_BAD_REQUEST
    assert login_response.json()["detail"] == "Invalid password"


async def test_account_me(
    http_api_client: AsyncTestClient, auth_service: AuthService, activated_data: ActivatedDataFixture
):
    login_data = await auth_service.login(
        LoginRequestDTO(
            email=activated_data.user.email,
            password=activated_data.user.password,
            device="testclient",
        ),
    )

    account_response = await http_api_client.get(
        "/api/auth/account/me",
        headers={API_KEY_HEADER: f"{TOKEN_PREFIX} {login_data.token}"},
    )
    assert account_response.status_code == HTTP_200_OK
    assert activated_data.user.email == account_response.json()["user"]["email"]


async def test_account_logout(http_api_client: AsyncTestClient, auth_service: AuthService, activated_data):
    login_data = await auth_service.login(
        LoginRequestDTO(email=activated_data.user.email, password=activated_data.user.password, device="testclient")
    )
    await http_api_client.post(
        "/api/auth/account/logout",
        headers={API_KEY_HEADER: f"{TOKEN_PREFIX} {login_data.token}"},
    )
    account_response = await http_api_client.get(
        "/api/auth/account/me",
        headers={API_KEY_HEADER: f"{TOKEN_PREFIX} {login_data.token}"},
    )
    assert account_response.status_code == HTTP_401_UNAUTHORIZED


async def test_verify_account_email(http_api_client: AsyncTestClient, auth_service: AuthService):
    user = UserFactory.build()
    auth_code = await auth_service.signup(SignUpRequestDTO(email=user.email, password=user.password))
    activate_response = await http_api_client.post("/api/auth/account/activate", json={"code": auth_code.code})
    assert activate_response.status_code == HTTP_201_CREATED

    login_data = await auth_service.login(
        LoginRequestDTO(
            email=user.email,
            password=user.password,
            device="testclient",
        )
    )

    account_data = await auth_service.get_account(login_data.token)
    assert account_data.user.email == user.email


async def test_not_authorized_with_wrong_tokenkey(
    http_api_client: AsyncTestClient, auth_service: AuthService, activated_data: ActivatedDataFixture
):
    login_data = await auth_service.login(
        LoginRequestDTO(
            email=activated_data.user.email,
            password=activated_data.user.password,
            device="testclient",
        )
    )

    account_response = await http_api_client.get(
        "/api/auth/account/me",
        headers={API_KEY_HEADER: f"{TOKEN_PREFIX} {login_data.token}_invalid"},
    )
    assert account_response.status_code == HTTP_401_UNAUTHORIZED
